////////////////////////////////////////////////////////////////
// Copyright (C) 1992, 1993, 1994 Sheetal Kakkad, Donovan Kolbly and
// Paul Wilson
//
// This file is part of the Texas Persistent Storage system.
//
// The Texas Persistent Storage system is free software; you can
// redistribute it and/or modify it under the terms of the GNU General
// Public License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// The Texas Persistent Storage system is distributed in the hope that
// it will be useful, but WITHOUT ANY WARRANTY; without even the
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE. See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with the Texas Persistent Storage system; if not, write to
// the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA
// 02139, USA.
//
// You can contact the authors by email at <oops@cs.utexas.edu>.
//
// Code modified by Adel Shrufi (a couple of fixes, and cursors).

#include <assert.h>

#ifndef BTREE_HH
#include "btree.hh"
#endif

// Regular binary search, with a hack that if a key is not found, the
// index of the next higher key in the array.
template <class K, class V>
int BTNode<K,V>::binary_search (const K& key, int &found) const
{
  int low, high;

  low = 0;
  high = count - 1;
  while (low <= high)
  {
    int mid = (low + high) / 2;
    
    if (keys[mid] < key)
      low = mid + 1;
    else
      if (keys[mid] > key)
	high = mid - 1;
      else
      {
	assert(keys[mid] == key);
	found = 1;
	return mid;
      }
  }

  // At this point, we know that the key is not found, and so we just
  // look for the index of the next higher key, and return it.
  found = 0;
  while ( (low < count) && (keys[low] < key) )
    low++;
  return low;
}


// return a pointer to the node and index within the node containing key
template <class K, class V>
const BTNode<K,V> *BTLeafNode<K,V>::search(const K& key, int& keyidx,
					   int& found) const
{
  // search always terminate in the leaf node
  found = 0;
  // Using linear search, need to replace it with binary search
#if 0
  for (int i = 0; i < count && key > keys[i]; i++)
    ;
  keyidx = i;
  if (i < count && key == keys[i])
    found = 1;
#endif
  keyidx = binary_search (key, found);

  return this;
}

// return a pointer to the leaf node where the key might be located
template <class K, class V>
const BTNode<K,V> *BTNonLeafNode<K,V>::search(const K& key, int& keyidx,
					      int& found) const
{
  // Using linear search, need to replace it with binary search
#if 0
  for (int i = 0; i < count && key > keys[i]; i++)
    ;
#endif
  int i = binary_search (key, found);
  // the left child of keys[i] is child[i],
  // it's right child is child[i+1]
  // the left subtree contains key <= keys[i]
  return child[i]->search(key, keyidx, found);
}

template <class K, class V>
void BTLeafNode<K,V>::insert_nonfull(const K& newkey, const V& newdata)
{
  // inserting into leaf node
  assert(!isfull());

  int found = 0;
  int i = binary_search (newkey, found);

  // the left child of keys[i] is child[i],
  // it's right child is child[i+1]
  // the left subtree contains key <= keys[i]

  // FIXME: what to do with duplicates? provide option?
#if 0
  if (newkey == keys[i])
    return;
#endif
  if (found)
  {
    assert(newkey == keys[i]);
    return;
  }

  // make room for newkey
  for (int j = count-1; j >= i; j--) 
  {
    keys[j + 1] = keys[j];
    values[j + 1] = values[j];
  }
  keys[i] = newkey;
  values[i] = newdata;
  count++;
}

template <class K, class V>
void BTNonLeafNode<K,V>::insert_nonfull(const K& newkey, const V& newdata)
{
  assert(!isfull());
  // determine the child node to descend into
  int found = 0;
  int i = binary_search (newkey, found);
  assert(!found);  // We forbid duplicate keys.

#if 0
  for (int i = 0; i < count && newkey > keys[i]; i++)
    ;
  // FIXME: what to do with duplicates? provide option?
  if (newkey == keys[i])
    return;
#endif
  // the left child of keys[i] is child[i],
  // it's right child is child[i+1]
  // the left subtree contains key <= keys[i]

  // if the child we're going to descend into is full, we split it
  if (child[i]->isfull()) 
  {
    child[i]->split(this, i);
    if (newkey > keys[i])	// FIXME: is this right???
      i++;
  }
  child[i]->insert_nonfull(newkey, newdata);
}

template <class K, class V>
void BTLeafNode<K,V>::split(BTNode<K,V> *the_parent, const int i)
{
  typedef BTLeafNode<K,V> BTLeafNodeKV;

  BTNonLeafNode<K,V> *parent = (BTNonLeafNode<K,V> *) the_parent;
  assert(!the_parent->isleaf());

  // *this is a full, i-th child of *parent.
  // *this is split in two about its median key, so that *parent
  // has an additional child.
  // parent is always a non-leaf node

  // create the same type of node as the one being split
  // BTLeafNode<K,V> *newnode = new BTLeafNode<K,V>;
#if defined(TEXAS) && defined(USE_INDEX)
  BTLeafNode<K,V> *newnode = pnew (obj_store_of (this), BTLeafNodeKV);
#else
  BTLeafNode<K,V> *newnode = new BTLeafNodeKV;
#endif
  nextleaf = newnode;	// chain in with this one

  // compute index to median key in *this
  int mid = (maxkeys - 1) / 2;
  
  // move elements > median key to newnode
  for (int j = mid + 1, k = 0; j <= count-1; k++, j++) 
  {
    newnode->keys[k] = keys[j];
    newnode->values[k] = values[j];
    newnode->count++;
  }
  // newnode->count = count - mid - 1; // XX: off by 1?
  // median key is retained in this node
  count -= newnode->count;

  // Splice in this node's median key to parent node
  parent->splice_key (i, keys[mid], newnode);
}


template <class K, class V>
void BTNonLeafNode<K,V>::splice_key (int index, const K& k, BTNode<K,V> *node)
{
  for (int i = count - 1; i >= index; i--) 
  {
    keys[i + 1] = keys[i];
    child[i + 2] = child[i + 1];
  }
  keys[index] = k;
  child[index+1] = node;
  count++;
}

template <class K, class V>
void BTNonLeafNode<K,V>::split(BTNode<K,V> *the_parent, const int i)
{
  typedef BTNonLeafNode<K,V> BTNonLeafNodeKV;

  BTNonLeafNode<K,V> *parent = (BTNonLeafNode<K,V> *) the_parent;
  assert(!the_parent->isleaf());

  // *this is a full, i-th child of *parent.
  // *this is split in two about its median key, so that *parent
  // has an additional child.
  // parent is always a non-leaf node

  // create the same type of node as the one being split
  // BTNonLeafNode<K,V> *newnode = new BTNonLeafNode<K,V>;
#if defined(TEXAS) && defined(USE_INDEX)
  BTNonLeafNode<K,V> *newnode = pnew (obj_store_of (this), BTNonLeafNodeKV);
#else
  BTNonLeafNode<K,V> *newnode = new BTNonLeafNodeKV;
#endif

  // compute index to median key in *this
  int mid = (maxkeys - 1) / 2;

  // move elements > median key to newnode
  newnode->child[0] = child[mid + 1];  // XXX: ???
  for (int j = mid + 1, k = 0; j <= count-1; k++, j++) 
  {
    newnode->keys[k] = keys[j];
    newnode->child[k] = child[j];
    newnode->count++;
  }
  newnode->child[k] = child[j]; // # of child is (# of keys + 1)
  
  // newnode->count = count - mid - 1;
  // the median key moves to the parent
  count -= (newnode->count + 1);

  // Splice in this node's median key to parent node
  parent->splice_key (i, keys[mid], newnode);
}

//
// ------------------------------- BTree ---------------------------------
//

// return a pointer to data, or nil if not found
template <class K, class V>
const V *BTree<K,V>::lookup(const K& key)
{
  int idx, found;
  const BTNode<K,V> *node = root->search(key, idx, found);

  assert(node->isleaf());
  if (found)
    // XXX: downcast is unsafe, should use dynamic_cast<> operator
    return &((BTLeafNode<K,V> *) node)->values[idx];
  else
    return 0;
}

template <class K, class V>
void BTree<K,V>::remove(const K& key)
{
  // Not implemented
}

template <class K, class V>
void BTree<K,V>::insert(const K& key, const V& data)
{
  typedef BTNonLeafNode<K,V> BTNonLeafNodeKV;

  // pre: visited node is non-full, or split it if it's full
  if (root->isfull()) 
  {
    // root node is full, so split it.
    // splitting the root is the only way to increase the height
    // of a B tree.
    // BTNonLeafNode<K,V> *newnode = new BTNonLeafNode<K,V>;
#if defined(TEXAS) && defined(USE_INDEX)
    BTNonLeafNode<K,V> *newnode = pnew (obj_store_of(this),BTNonLeafNodeKV);
#else
    BTNonLeafNode<K,V> *newnode = new BTNonLeafNodeKV;
#endif
    
    // previous root node becomes the left child of the new root node
    newnode->child[0] = root;
    
    // newnode is now the root
    root = newnode;
    (newnode->child[0])->split(newnode, 0);
  }
  root->insert_nonfull(key, data);
}

template <class K, class V>
BTCursor<K,V> *BTree<K,V>::cursor_of() const
{
  typedef BTCursor<K,V> BTCursorKV;

  return new BTCursorKV(this);
}

template <class K, class V>
const V* BTCursor<K,V>::Get_Value()
{
  if (_current_index < _current_LN->count) 
    return &_current_LN->values[_current_index];
  else
    return (V*) 0;
}
  
template <class K, class V>
const V* BTCursor<K,V>::first()
{ 
  First_Called = 1;  
  _current_index = 0;
  _current_LN = 0;
  const BTNode<K,V>  *Bnode = BT->root;

  // move down the leftmost child till we reach a leaf page
  while( !Bnode->isleaf() )
    Bnode = ((BTNonLeafNode<K,V> *) Bnode)->child[0];

  // make sure we're at a leaf node
  assert(Bnode->isleaf());

  _current_LN = (BTLeafNode<K,V>*) Bnode;

  //call next to return the first value
  return next();
}

// note that next return NULL only when the entire leaf page chain is
// traversed. This may not be needed for our situations since we never
// delete from the B-tree, but in general B-trees may have empty leaf
// pages reslting from deletes.

template <class K, class V>
const V* BTCursor<K,V>::next()
{
  assert(First_Called);
  const V* val = Get_Value();
  if (!val){ // if val is NULL, we check the next leaf page
    // move to the next leaf node if it exists
    _current_LN = _current_LN->nextleaf;
    if (_current_LN) { // make sure it is nonNULL 
      // reset the current_index and call next again
      _current_index = 0;
      return next();
    } else
      // no more leaf pages to traverse, so we're done!
      return (V*) 0;
  }
  // increment the current index
  _current_index++;
  return val;
}
