////////////////////////////////////////////////////////////////
// Copyright (C) 1992, 1993, 1994 Sheetal Kakkad, Donovan Kolbly and
// Paul Wilson
//
// This file is part of the Texas Persistent Storage system.
//
// The Texas Persistent Storage system is free software; you can
// redistribute it and/or modify it under the terms of the GNU General
// Public License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// The Texas Persistent Storage system is distributed in the hope that
// it will be useful, but WITHOUT ANY WARRANTY; without even the
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE. See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with the Texas Persistent Storage system; if not, write to
// the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA
// 02139, USA.
//
// You can contact the authors by email at <oops@cs.utexas.edu>.
//
// Code modified by Adel Shrufi (a couple of fixes, and cursors).

#ifndef BTREE_HH
#define BTREE_HH

/***********
   This package implements a b+ tree. We are not doing any rebalancing
   during btree:: remove() and it will create a problem when a node is
   empty but not deleted.  Dont do remove if you can and you should
   implement rebalancing to fix the problem.  Note that each node has
   m+1 slots for children and keys, instead of m. This is a hack to
   make it easier to do a split.  We still enforce that a node should
   have at most m children after an insert.
************/

#define PAGE_SIZE		(4096)
#define NODESIZE		(PAGE_SIZE - 128)

// forward declaration
template <class K, class V> class BTCursor;
template <class K, class V> class BTree;
template <class K, class V> class BTLeafNode;
template <class K, class V> class BTNonLeafNode;


template <class K, class V> class BTNode 
{
 protected:
  int count;

  enum { maxkeys = NODESIZE / (sizeof(K)+sizeof(V)) };

  K keys[maxkeys];
  
 public:
  friend class BTree<K,V>;
  friend class BTCursor<K,V>;

 protected:
  BTNode() { count = 0; }
  virtual ~BTNode() { }

  int binary_search (const K& key, int &found) const;

  int isfull() const { return count == maxkeys; };
  virtual int isleaf() const = 0;
  virtual const BTNode<K,V> *search (const K& k, int& keyidx, 
				     int& found) const = 0;
  virtual void insert_nonfull(const K& newkey, const V& newdata) = 0;
  // virtual void split(BTNonLeafNode<K,V> *, const int) = 0;
  virtual void split(BTNode<K,V> *, const int) = 0;
};

template <class K, class V> class BTLeafNode : public BTNode<K,V>
{
 private:
#if 0
  // Since there's a linear order on the leaves based upon
  // the search-key values they contain, we use a pointer to
  // chain the leaf nodes together in search-key order.
#ifdef SMARTINDEX
  Ptr<BTLeafNode<K,V> > nextleaf;
#else
  BTLeafNode<K,V> *nextleaf;
#endif
#endif

  V values[maxkeys + 1];
  BTLeafNode<K,V> *nextleaf;

 protected:
  friend BTree<K,V>;
  friend BTCursor<K,V>;

  BTLeafNode() : nextleaf (NULL) { }
  ~BTLeafNode() { }

  const BTNode<K,V> *search(const K& key, int& keyidx, int& found) const;
  int isleaf() const { return 1; }
  void insert_nonfull(const K& newkey, const V& newdata);
  // void split(BTNonLeafNode<K,V> *, const int);
  void split(BTNode<K,V> *, const int);
};

template <class K, class V> class BTNonLeafNode : public BTNode<K,V> 
{
 private:
  // internal node has pointers to its children
#ifdef SMARTINDEX
  Ptr<BTNode<K,V> > child[maxkeys+1];
#else
  BTNode<K,V> *child[maxkeys + 1];
#endif

 protected:
  friend class BTree<K,V>;
  friend class BTLeafNode<K,V>;
  friend BTCursor<K,V>;

  BTNonLeafNode() { }
  ~BTNonLeafNode() { }

  const BTNode<K,V> *search (const K& key, int& keyidx, int& found) const;
  int isleaf() const { return 0; }
  void insert_nonfull (const K& newkey, const V& newdata);
  // void split(BTNonLeafNode<K,V> *, const int);
  void split (BTNode<K,V> *, const int);
  void splice_key (int index, const K& key, BTNode<K,V> *node);
};

template <class K, class V> class BTree 
{
 private:
#if defined (SMARTINDEX) && defined (ALLSMARTINDEX)
  Ptr<BTNode<K,V> > root;
#else
  BTNode<K,V> *root;
#endif

  typedef BTLeafNode<K,V> BTLeafNodeKV;
  typedef BTNonLeafNode<K,V> BTNonLeafNodeKV;

 public:
#if defined(TEXAS)
  BTree() { root = pnew (obj_store_of (this), BTLeafNodeKV); }
#else
  BTree() { root = new  BTLeafNodeKV; }
#endif
  ~BTree() { delete root; }
  const V *lookup (const K& key);
  void insert (const K& key, const V& data);
  void remove (const K& key);
  friend class BTCursor<K,V>;
  BTCursor<K,V> *cursor_of() const; 
};

template <class K, class V> class BTCursor
{
protected:
  short current_index() const { return _current_index; }
  short _current_index;
  BTLeafNode<K,V>  *_current_LN;  // current leaf node
  short First_Called;  // ensures that first() was called first 
  const BTree<K,V>  *BT;  // The B-tree being iterated over
  const V* Get_Value();
public:
  BTCursor(const BTree<K,V> *bt1):_current_index(0),_current_LN(0),
			    BT(bt1), First_Called(0) {}
  const V*   first();
  const V*   next();
};



#ifdef __GNUG__
#include "btree.cc"
#endif

#endif
