/* 
 * Copyright (c) 1995
 *       Whitehead Institute for Biomedical Research. All rights reserved.
 *
 * Released under terms specified in the Copyright file present in the top
 * directory of this distribution.
 */


// $Id: builtin1.C,v 1.6 1995/02/13 17:49:10 steve Exp $
#include "builtin1.H"

#ifndef __STRING_H
#include <string.h>
#endif

#ifndef BASICS_H
#include "basics.H"
#endif

#ifndef QL_ARGS_H
#include "ql_args.H"
#endif

#ifndef QL_EVAL_H
#include "ql_eval.H"
#endif

#ifndef QL_ITERATOR_H
#include "ql_iterator.H"
#endif

#ifndef QL_ERROR_H
#include "ql_error.H"
#endif

#ifndef VALUE_H
#include "value.H"
#endif

#ifndef QL_DD_H
#include "ql_dd.H"
#endif

#ifndef QL_DD_ENTRY_H
#include "ql_dd_entry.H"
#endif

#ifndef STEP_H
#include "step.H"
#endif

#ifndef MATERIAL_H
#include "material.H"
#endif

#ifndef LIST_H
#include "list.H"
#endif

static QL_Term *
gather_m_args_into_term_list(QL_Term *t, int m, int*error_found)
// Take the initial "m" arguments from t, and term them into a term list
// (usually for evaluation).
//
// The first m arguments of "t" must be terms (and be bound),
// or an error is generated.
// "*error_found" is set to 0 if everything is ok, otherwise to 1.
{
  *error_found = 0;
  QL_Term *arg_terms = 0, *q = 0;
  QL_Args *args = t->args();
  char tmp[MAX_ARITY + MAX_ARITY];

  for (int i = 0; i < m; i++, args = args->next()) {
    if (!args->value()) {
      sprintf(tmp,"%d",m);
      ql_eval_error(t, "The first ", tmp, " arguments must be bound in");
      *error_found = 1;
      return 0;
    }
    QL_Term *t2 = tv2term(t, args->value());
    if (!t2) {
      sprintf(tmp,"%d",m);
      ql_eval_error(t, "Argument ", tmp, " must be a term in");
      *error_found = 1;
      return 0;
    }
    if (!q) arg_terms = t2;
    else {
      if (q->next() && args->value() == t2)
	{ ABORT(q->next() != t2, "Problem linking term args"); }
      else { q->set_next(t2); }
    }	   
    q = t2;
  }
  return arg_terms;
}


QL_FIRST_HEADER(count_pred)
{
  if (in); // Suppress warnings.
  if (n != 1) {
    ql_eval_error(t, "Require exactly the last argument to be free in");
    return 0;
  }
  int error_found;
  QL_Term *arg_terms = gather_m_args_into_term_list(t, m, &error_found);
  if (error_found) return 0;
  int count = arg_terms ? QL_eval_no_print(arg_terms) : 0;
  clear_error();
  out[0] = Long_Int::make(count);
  if (errmsg()) {
    ql_eval_error(t, errmsg());
    clear_error();
    return 0;
  }
  return (QL_Iterator_State*) 1;
}


static List_Value*
new_list() {
  return new List_Value;
}


static void
destroy_list(const void *list)
{
  delete (List_Value*)list;
}


static List_Value*
add_to_list(QL_Term *, List_Value* x, Value** elements, int n)
{
  ABORT(n != 1, "Need exactly 1 tag value");
  x->put(elements[0]);
  return x;
}

// Add to a Value_List, but do not insert duplicate values.
static List_Value* 
add_to_set(QL_Term *t, List_Value* x, Value** elements, int n)
{
  ABORT(n != 1, "Need exactly 1 tag value");
  Cursor_Value *c = x->cursor_of();
  for (Value *v = c->first(); v; v = c->next()) {
    clear_error();
    int contained =  (*v == *elements[0]);
    if (errmsg()) {
      ql_eval_error(t, errmsg(), " in");
      delete c;
      return 0;
    }
    if (contained) { delete c; return x; }
  }
  delete c;
  x->put(elements[0]);
  return x;
}


static QL_Iterator_State *
gather_in_any_pred(QL_Term*t,
		   int in_number, Value*in[], int out_number, Value* out[],
		   List_Value *(*init_fn)(),
		   void (*destroy_fn)(const void *),
		   List_Value *(*add_fn)(QL_Term*, List_Value*, Value**, int),
		   int check_result_address, Value::tv_type_id expected_type)
{
  if (in); // Supress warning;
  if (out_number != 2) {
    ql_eval_error(t, "Require exactly the last two arguments to be free in");
    return 0;
  }
  if (in_number < 1) {
    ql_eval_error(t, "Require at least one term argument in");
    return 0;
  }
  int error_found;
  QL_Term *term_list = gather_m_args_into_term_list(t, in_number, &error_found);
  if (error_found) return 0;
  void *result = init_fn();
  QL_Binding_Node *binding_nodes[1];

  QL_Args *arg = t->args();

  for (int i = in_number; i; i--) {
    ABORT(!arg, "args should not be 0");
    arg = arg->next();
  }
  binding_nodes[0] = arg->node();
  ABORT(!binding_nodes[0], "Must have a variable here");
  ABORT(binding_nodes[0]->value(), "Variable must be unbound here");
  if (binding_nodes[0] == arg->next()->node()) {
    destroy_fn(result);
    ql_eval_error(t, "Variable ", binding_nodes[0]->identifier(), 
		  " cannot appear as both the last and next-to-last argument of");
    return 0;
  }

  {
    void *temp
      = QL_eval_with_aggregate(t, term_list, result, binding_nodes,
			       out_number - 1, 
			       (void *(*)(QL_Term *, void *, Value **, int)) add_fn);
    if (!temp) {
      destroy_fn(result);
      return 0;
    }
    ABORT(check_result_address && result != temp,
	  "We lost the address of 'result'");
  }
  out[out_number-1] 
	  = V_Collection::make(expected_type, (List_Value*)result);
  destroy_fn(result);
  return (QL_Iterator_State*)1;
}


QL_FIRST_HEADER(gather_in_list_pred)
{
  return gather_in_any_pred(t, m, in, n, out, new_list, destroy_list,
			    add_to_list,1, Value::LIST);
}


QL_FIRST_HEADER(gather_in_set_pred)
{
  return gather_in_any_pred(t, m, in, n, out, new_list, destroy_list,
			    add_to_set, 1, Value::SET);
}


QL_FIRST_HEADER(not_pred)
{
  if (out || in); // Suppress the warning;
  if (0 != n) {
    ql_eval_error(t, "All arguments must be bound in");
    return 0;
  }
  if (0 == m) return 0;
  int error_found;
  QL_Term *arg_terms = gather_m_args_into_term_list(t, m+n, &error_found);
  if (error_found) return 0;
  if (QL_eval_no_print(arg_terms)) return 0;
  else return (QL_Iterator_State*)1;
}


QL_FIRST_HEADER(insist_pred_first)
{
  abort_m_n(m,1,n,0);
  if (out); // Suppress warning.
  QL_Term *arg_term = tv2term(t,in[0]);
  if (!arg_term) {
    ql_eval_error(t, "Expect a single term as argument to");
    return 0;
  }
  QL_Iterator *iterator = create_iterator(arg_term);
  if (iterator && iterator->true_bindings())
    return (QL_Iterator_State*)iterator;
  else {
    if (ql_eval) ql_eval_error(arg_term,"No true bindings for");
    delete iterator;
    return 0;
  }
}


QL_NEXT_HEADER(insist_pred_next)
{
  if (in || out) m = 0; n = 0;  t = 0; // Suppress warnings.
  QL_Iterator *iterator = (QL_Iterator *) s;
  if (iterator->true_bindings()) return s;
  else {
    delete iterator;
    return 0;
  }
}


static int is_no_comparison(QL_Term *t, Value* v)
{
  if (v->tv_type() == Value::SET) {
    ql_eval_error(t, "First argument cannot be a 'SET' for");
    return 1;
  } else return 0;
}


enum relation_op {gt, lt, ge, le, ne, eq};


static int check_relation_op_type(QL_Term *t, Value*tv0, Value*tv1,
				  relation_op op,
				  Value::tv_type_id tv_type)
{
  if ((tv0->tv_type() == tv_type || tv1->tv_type() == tv_type)
      && (op != eq && op != ne)) {
    ql_eval_error
      (t, "Only '==' and '<>' are legal comparisons for a '",
       Value::type2char(tv_type), "' in");
    return 0;
  } 
  return 1;
}

static int
any_relation_op(QL_Term*t, int m, Value*a[], int n, relation_op op)
{
  abort_m_n(m,2,n,0);
  Value *tv0 = a[0], *tv1 = a[1];
  if (is_no_comparison(t, tv0) || is_no_comparison(t, tv1)) return 0;
  if (!check_relation_op_type(t, tv0, tv1, op, Value::STEP_POINTER))
    return 0;
  if (!check_relation_op_type(t, tv0, tv1, op, Value::MATERIAL_POINTER))
    return 0;
  if (!check_relation_op_type(t, tv0, tv1, op, Value::TERM))
    return 0;

  clear_error();
  int r = 0;
  switch (op) {
  case gt: r = (*tv0 > *tv1); break;
  case lt: r = (*tv0 < *tv1); break;
  case ge: r = (*tv0 >= *tv1); break;
  case le: r = (*tv0 <= *tv1); break;
  case ne: r = (*tv0 != *tv1); break;
  case eq: r = (*tv0 ==*tv1); break;
  default: ABORT(1,"illegal boolean operator");
  }
  if (errmsg()) { ql_eval_error(t, errmsg(), " in"); return 0; }
  return r;
}


#define RELOP(OP) \
QL_Iterator_State *CAT_ID2(OP,_pred)(QL_Term*t, int m, Value*a[],\
                                int n, Value*[]) \
{ \
  return (QL_Iterator_State*) any_relation_op(t, m, a, n, OP); \
}

RELOP(gt)
RELOP(lt)
RELOP(le)
RELOP(ge)
RELOP(ne)
RELOP(eq)

#undef RELOP


QL_FIRST_HEADER(next_pred)
{
  abort_m_n(m,3,n,0);
  if(out); // To suppress warning.
  if (in[0]->tv_type() != Value::MATERIAL_POINTER) {
    ql_eval_error(t, "Expect 'MATERIAL_POINTER' as type of first argument of");
    return 0;
  }
  if (in[1]->tv_type()!=Value::STRING
      || in[2]->tv_type()!=Value::STRING) {
    ql_eval_error(t,
		  "Expect 'STRING' as type of second and third arguments of");
    return 0;
  }

  char *step_kind1 = in[1]->down_to_STRING()->get(),
       *step_kind2 = in[2]->down_to_STRING()->get();

  QL_DD_Entry *e1 = QL_DD::find(step_kind1), *e2 = QL_DD::find(step_kind2);
  if (!e1 || e1->entry_kind() != QL_DD_Entry::STEP_KIND) {
    ql_eval_error(t, "'", step_kind1, "' is not a step kind in");
    delete step_kind1; delete step_kind2;  return 0;
  }
  if (!e2 || e2->entry_kind() != QL_DD_Entry::STEP_KIND) {
    ql_eval_error(t, "'", step_kind2, "' is not a step kind in");
    delete step_kind1; delete step_kind2;  return 0;
  }
  delete step_kind1; delete step_kind2;

  if (e1 == e2) {
    ql_eval_error(t, "Second and third arguments must be different in");
    return 0;
  }

  Cursor_Step *c = in[0]->down_to_MATERIAL_POINTER()->step_cursor();

  for (Step *s = c->last(); s; s = c->previous()) {
    if (s->step_kind() == e2) { delete c; return 0; } // s2 follows s1
    else if (s->step_kind() == e1) { // s1 without a following s2
      time_t d = s->timestamp();
      while ((s = c->previous()) && d == s->timestamp())
	if (s->step_kind() == e2) { delete c; return 0; }
      delete c; return (QL_Iterator_State*)1;
    }
  }
  delete c; return 0;
}


QL_FIRST_HEADER(substring_bbbf_pred)
{
  abort_m_n(m,3,n,1);
  Value *string = in[0], *start = in[1], *length = in[2];
  String *sv;
  if (string->tv_type() != Value::STRING
	  && string->tv_type() != Value::DNA_SEQUENCE) {
    ql_eval_error
      (t, "First argument must be of type 'STRING' or 'DNA_SEQUENCE' in");
    return 0;
  } else sv = string->down_to_STRING();

  if (start->tv_type() != Value::INTEGER
      || length->tv_type() != Value::INTEGER) {
    ql_eval_error
      (t, "Arguments 2 and 3 must be of type 'INTEGER' in");
    return 0;
  }

  clear_error();
  out[0] = sv->substring(int(*start->down_to_INTEGER()),
			 int(*length->down_to_INTEGER()));
  if (errmsg()) { ql_eval_error(t, errmsg(), " in"); return 0; }
  if (!out[0]) { ql_eval_error(t, "Possible programming error in"); return 0; }
  return (QL_Iterator_State*) 1;
}


QL_FIRST_HEADER(reversec2_pred)
{
  abort_m_n(m,1,n,1);
  Value *tv = in[0];
  Sequence *seq = 0, *in_seq;
  if (tv->tv_type() == Value::DNA_SEQUENCE)
    in_seq = tv->down_to_DNA_SEQUENCE();
  else if (tv->tv_type() == Value::STRING)
    in_seq = seq = Sequence::make(tv->down_to_STRING());
  if (!in_seq->valid()) { 
    ql_eval_error
      (t, "'STRING' argument cannot be converted to 'DNA_SEQUENCE' in");
    return 0;
  }
  
  clear_error();
  Sequence *r = in_seq->reverse_complement();
  if (errmsg()) { ql_eval_error(t, errmsg(), " in"); return 0; }
  if (!r) { ql_eval_error(t, "Possible programming error in"); return 0; }
  out[0] = r;
  return (QL_Iterator_State*)1;
}

// Not used..
//QL_FIRST_HEADER(reversec_bbbf_pred)
//{
// if (in || out) // Suppress warnings.
//  abort_m_n(m,3,n,1);
//  ql_eval_error(t, "No implementation for");
//  return 0;
//}


#define TYPE_ERROR \
  {ql_eval_error(t, "The arguments must be 'LIST'('INTEGER'), 'INTEGER', 'INTEGER', 'INTEGER' for"); \
  return 0;}


QL_Iterator_State*
either_polymorphic_pred(QL_Term *t, Value*in[], Value*out[],
			int strict)
{
  if (out); // Suppress warnings
  Value *sizes_tv=in[0], *i1_tv=in[1], *i2_tv=in[2], *epsilon_tv=in[3];
  if (sizes_tv->tv_type() != Value::LIST
      || i1_tv->tv_type() != Value::INTEGER
      || i2_tv->tv_type() != Value::INTEGER
      || epsilon_tv->tv_type() != Value::INTEGER)
    TYPE_ERROR;
  V_Collection *list = sizes_tv->down_to_LIST();
  int card = list->cardinality();
  int index1 = int(*i1_tv->down_to_INTEGER()),
      index2 = int(*i2_tv->down_to_INTEGER());
  if (index1 < 0 || index1 >= card) {
    ql_eval_error(t, "Second argument is out of bounds in");
    return 0;
  }
  if (index2 < 0 || index2 >= card) {
    ql_eval_error(t, "Third argument is out of bounds in");
    return 0;
  }
  Value *tv1 = list->get(index1);
  Value *tv2 = list->get(index2);
  if (tv1->tv_type() != Value::INTEGER 
      || tv1->tv_type() != Value::INTEGER)
    TYPE_ERROR;
  int size1 = int(*tv1->down_to_INTEGER());
  int size2 = int(*tv2->down_to_INTEGER());
  int e = int(*epsilon_tv->down_to_INTEGER());
  if (strict && (size1 == -1 || size2 == -1)) return 0;
  if (!strict && size1 == -1 && size2 == -1) return 0;
  if (!strict && (size1 == -1 || size2 == -1))
     return (QL_Iterator_State*) 1;
  if ((size1 >= size2)  && ((size1 - size2) < e)) return 0;
  if ((size2 > size1)  && ((size2 - size1) < e)) return 0;
  return (QL_Iterator_State*) 1;
}
#undef TYPE_ERROR


QL_FIRST_HEADER(polymorphic_pred)
{
  abort_m_n(m,4,n,0);
  return either_polymorphic_pred(t, in, out, 0);
}


QL_FIRST_HEADER(strictly_polymorphic_pred)
{
  abort_m_n(m,4,n,0);
  return either_polymorphic_pred(t, in, out, 1);
}


QL_FIRST_HEADER(true_pred)
{
  abort_m_n(m,0,n,0);
  if(in || out || t); // Suppress warnings.
  return (QL_Iterator_State*)1;
}

//QL_FIRST_HEADER(positive_pred)
//{
 // abort_m_n(m,0,n,0);
  //if(in || out || t); // Suppress warnings.
  //return (QL_Iterator_State*)1;
//}

QL_FIRST_HEADER(var_pred)
{
  if (m + n != 1) {
    ql_eval_error(t, "Expect exactly 1 argument for");
    return 0;
  }
  if (in || out); // Suppress compiler warning.
  if (t->args()->value()) return 0;
  else return (QL_Iterator_State*)1;
}


QL_FIRST_HEADER(exists_pred)
{
  if (out || in); // Suppress the warning;
  if (0 != n) {
    ql_eval_error(t, "All arguments must be bound in");
    return 0;
  }
  if (0 == m) return 0;
  int error_found;
  QL_Term *arg_terms = gather_m_args_into_term_list(t, m+n, &error_found);
  if (error_found) return 0;
  if (QL_eval_exists(arg_terms)) return (QL_Iterator_State*)1;
  else return 0;
}


static Value::tv_type_id or_types[] = { Value::TERM, Value::TERM };
BUILTIN_FIRST_DEFINITION(or_pred_first,2,0,or_types,2,1)
{
  if(out_tv || m || n); // Suppress warnings.
  // If the first term succeeds
  // then succeed once for each time
  // the first term succeeds.  Otherwise
  // succeed once for each time the second
  // term succeeds.  (Similar to insist.)
  QL_Term *t0 = tv2term(t,in_tv[0]), *t1 = tv2term(t,in_tv[1]);
  if (!t0 || !t1) {
    ql_eval_error(t, "Expect two terms to be arguments to");
    return 0;
  }
  QL_Iterator *iterator = create_iterator(t0);
  if (iterator && iterator->true_bindings())
    return (QL_Iterator_State*)iterator;
  else {
    delete iterator;
    if (ql_eval) {
      iterator = create_iterator(t1);
      if (iterator && iterator->true_bindings())
	return (QL_Iterator_State*)iterator;
      else delete iterator;
    }
  }
  return 0;
}
