/* 
 * Copyright (c) 1995
 *       Whitehead Institute for Biomedical Research. All rights reserved.
 *
 * Released under terms specified in the Copyright file present in the top
 * directory of this distribution.
 */


#include "collated_tags.H"

#ifndef VALUE_H
#include "value.H"
#endif

#ifndef QL_ARGS_H
#include "ql_args.H"
#endif

#ifndef QL_ERROR_H
#include "ql_error.H"
#endif

#ifndef QL_DD_ENTRY_H
#include "ql_dd_entry.H"
#endif

#ifndef QL_TERM_H
#include "ql_term.H"
#endif

#ifndef MATERIAL_H
#include "material.H"
#endif

#ifndef ID_TAG_ELEMENT_H
#include "id_tag_element.H"
#endif

// Collated_Tags::ct_entry Collated_Tags::_entries[];

Collated_Tags *Collated_Tags::collate(QL_Term *t,QL_Args *args, int dd_tags_ok)
{
  static Collated_Tags r;
  r._tag_number = 0; // clear the table;
  if (r._ids) {  // clear the set of ids;
    Cursor_Id_Tag_Element *delete_cursor = r._ids->cursor_of();
    for (Id_Tag_Element* te = delete_cursor->first();
	 te; te = delete_cursor->next()) delete te;
    delete delete_cursor;
    delete r._ids;
  }
  r._ids = new List(Id_Tag_Element)(2);
  if (r._materials) { // clear the set of materials;
    delete r._materials; 
  }
  r._materials = new List_Material(3);

  // THIS CODE NEES TO BE MODIFIED to accommodate the
  // explicit ``associate_with_id'' attribute of
  // tag entries.  It should be an error if there
  // is an id tag, and no material in the ``id_materials''
  // set.  This is probably simplest to implement in two
  // passes over the arguments: pass 1 gets the id_materials
  // and the history_materials, and pass 2 makes neccessary
  // entries in the id indexes.

  for (QL_Args *p = args; p; p = p->next()) { if (!r.add_arg(t,p,dd_tags_ok)) return 0; }
  if (r._ids->count() > 0 && r._materials->count() > 1) {
    ql_eval_error(t, "Trying to associate an id with more than one material");
    return 0;
  }
  return &r;
}


inline void Collated_Tags::add_1_tag_value(ct_entry *e, Value *tv)
{
  e->v[e->count()] = tv;
  e->increment_count();
}


inline Collated_Tags::ct_entry *Collated_Tags::find_or_create(QL_Term *t, QL_DD_Tag_Entry *tag_entry)
{
  ct_entry *e = find(tag_entry->tag_index());
  if (e) {
    ql_eval_error(t, "Duplicate instances of tag '",
		    tag_entry->identifier(), "' in");
    return 0;
  }  else {
    e = &_entries[_tag_number]; _tag_number++;
    e->set_tag_entry(tag_entry);
    e->zero_count();
    return e;
  }
}


int Collated_Tags::add_arg(QL_Term *t, QL_Args *arg, int dd_tags_ok)
{
  // Arg has to be a QL_Tag_Assoc, otherwise it is a user error.
  QL_Tag_Assoc *t_a = arg->coerce_to_Tag_Assoc();
  if (!t_a) {
    ql_eval_error(t,"All arguments must be of the form tag=value in");
    return 0;
  }

  if (t_a->tag()->dd_tag() && !dd_tags_ok) {
    ql_eval_error(t, "Data dictionary tag '",
		  t_a->tag()->identifier(), "' is illegal in");
    return 0;
  }

  if (t_a->op() != QL_Tag_Assoc::EQUAL) {
				// OK until we allow updates!
    ql_eval_error(t, "Only = is a legal tag operator for");
    return 0;
  }

  Tag tag = t_a->tag()->tag_index();
  if (Tag::dd_type() == tag) return 1; // We skip this one because
                                       // its handling is unique.
				       // See QL_Type_Descriptor::find_in_term.

  Value *tv = t_a->check_or_coerce_type(t);
  if (!tv) return 0;
  QL_DD_Tag_Entry *tag_entry = t_a->tag();
  ct_entry *e = find_or_create(t, tag_entry);
  if (!e) return 0;
  if (!maybe_do_id_tag(t,tag_entry, tv)) return 0;
  if (tv->tv_type() == Value::MATERIAL_POINTER)
    _materials->insert(tv->down_to_MATERIAL_POINTER());
  add_1_tag_value(e,tv);
  return 1;
}
					
int Collated_Tags::maybe_do_id_tag(QL_Term *t, QL_DD_Tag_Entry *tag_entry,
				   Value *tv)
{
  if (tag_entry->id_tag()) {
    ABORT(Value::STRING != tv->tv_type(), "Non-string identifier not implemented");
    char *id = tv->down_to_STRING()->get();
    if (!tag_entry->check_duplicate(t, id)) { delete id; return 0; }
    else delete id;
    _ids->insert(new Id_Tag_Element(tag_entry,tv));
  }
}

int Collated_Tags::add_tag_value(QL_Term *t, QL_DD_Tag_Entry *tag_entry, Value *tv)
{
  ct_entry *e = find_or_create(t, tag_entry);
  if (!e) return 0;

  if (!maybe_do_id_tag(t,tag_entry, tv)) return 0;

  if (tv->tv_type() == Value::MATERIAL_POINTER)
    _materials->insert(tv->down_to_MATERIAL_POINTER());
  add_1_tag_value(e,tv);
  
  if (_ids->count() > 0 && _materials->count() > 1) {
    ql_eval_error(t, "Trying to associate an id with more than one material");
    return 0;
  }
  return 1;
}


int Collated_Tags::material_number() const
{
  return _materials->count();
}


Collated_Tags::ct_entry *Collated_Tags::find(Tag tag) const
{
  for (int i=0; i < tag_number(); i++)
    if (tag == _entries[i].tag_entry()->tag_index()) return &_entries[i];
  return 0;
}


int Collated_Tags::contain(Tag tag) const
{
  return find(tag) != 0;
}


int Collated_Tags::value_pointer_number() const
{
  int r=0;
  for (int i=0; i<tag_number(); i++) r += _entries[i].count();
  return r;
}


Value *Collated_Tags::find_tag_value(Tag tag) const
{
  ct_entry *e = find(tag);
  if (!e) return 0; else return e->v[0];
}
  

String *Collated_Tags::unique_string_value(Tag tag) const
{
  Value *v = find_tag_value(tag);
  ABORT(!v,"Cannot find tag");
  return v->down_to_STRING();
}


int Collated_Tags::unique_boolean_value(Tag tag) const
{
  ct_entry *e = find(tag);
  ABORT(!e,"Cannot find tag");
  ABORT(e->tag_entry()->type()->constructor() != Value::BOOLEAN,
	"Need a BOOLEAN");
  int r = int(*e->v[0]->down_to_INTEGER());
  ABORT(r < 0 || r > 1, "This cannot be a boolean");
  return r;
}


QL_DD_Tag_Entry *Collated_Tags::contain_create_only_tag() const
{
  for (int i=0; i < tag_number(); i++)
    if (_entries[i].tag_entry()->create_only()) return _entries[i].tag_entry();
  return 0;
}


int Collated_Tags::contain_id_material() const
{
  for (int i=0; i < tag_number(); i++)
    if (_entries[i].tag_entry()->put_on_history()
        && _entries[i].tag_entry()->type()->constructor()
	== Value::MATERIAL_POINTER) return 1;
  return 0;
}
