/* 
 * Copyright (c) 1995
 *       Whitehead Institute for Biomedical Research. All rights reserved.
 *
 * Released under terms specified in the Copyright file present in the top
 * directory of this distribution.
 */


%{
#ifndef QL_ERROR_H
#include "ql_error.H"
#endif

#ifndef QL_ARGS_H
#include "ql_args.H"
#endif

#ifndef BASICS_H
#include "basics.H"
#endif

#ifndef VALUE_H
#include "value.H"
#endif

#ifndef QL_TERM_H
#include "ql_term.H"
#endif;

#ifndef QL_EVAL_H
#include "ql_eval.H"
#endif

extern int yylex();
extern char ql_string[];

void yyerror(char *msg)
{
  ql_parse_error(msg);
}

static QL_Bindings *bindings = 0;

extern int yyparse();

extern int sigint_seen;  // Defined in main.C

int parse()
{
  bindings = new QL_Bindings;
  ql_eval = 1;
  sigint_seen = 0;
  // QL_eval is responsible for deleting the bindings
  // and any other temporarily allocated storage.
  // The production for the end of a query is
  // responsible for re-initializing bindings.
  
  Value::pushmem(); // Set a memory arena for Value's.
  int r = yyparse();
  Value::popmem_last(); // Reclaim the last of the memory.
  return r;
}

extern int respond_with_control_A;  // defined in ql.l

#define YYSTYPE void *
%}

%token DOT COMMA TWO_OP ERROR
%token L_PAREN R_PAREN L_LIST R_LIST L_SET R_SET L_BAG R_BAG
%token TOK_STRING TOK_DATE TOK_FLOAT TOK_INTEGER VARIABLE
%token WS_L_PAREN
// STRING, DATE, FLOAT, INTEGER renamed to avoid collision with
// tv_type_ids.

%start queries

%%

queries:         /* empty */
              | queries query
              ;

query:          terms DOT { QL_eval((QL_Term*)$1,bindings);
                            if (respond_with_control_A) cout << '\001' << flush;
                            delete bindings; bindings = new QL_Bindings;
                            QL_Term::gc(); QL_Args::gc();
                            Value::popmem_last(); Value::pushmem();
                            ql_eval = 1; sigint_seen = 0;}
              | DOT { yyerror("empty query");
                      if (respond_with_control_A) cout << '\001' << flush;
                      ql_eval = 1; sigint_seen = 0; }
              | error DOT { delete bindings; bindings = new QL_Bindings;
                            if (respond_with_control_A) cout << '\001' << flush;
                            QL_Term::gc(); QL_Args::gc();
                            Value::popmem_last(); Value::pushmem();
                            ql_eval = 1; sigint_seen =0; }
              ;

terms:          term { $$ = $1; }
              | term COMMA terms 
                { if ($1) ((QL_Term*)$1)->set_next((QL_Term*)$3);
		  $$ = $1;	  
                }
              ;

l_paren:       L_PAREN | WS_L_PAREN ;

term:            good_term {$$ = $1; }
               | simple_arg TWO_OP arg
                 { if ($1) { 
                      ((QL_Args*)$1)->set_next((QL_Args*)$3);
                      $$ = QL_Term::create((void*)$2,(QL_Args *)$1,
                                           error_position);
                   } else $$ = 0;
                 }
               | TWO_OP
                 { $$ = 0;
                   ql_parse_error
                     ("A binary operator cannot appear without arguments");
                 }
               | TWO_OP l_paren R_PAREN
                 { $$ = 0; 
                   ql_parse_error
                     ("A binary operator cannot appear without arguments");
                 }
               | bad_term  { $$ = $1; }
               ;

var_paren_term:    VARIABLE l_paren R_PAREN { $$ = 0; }
                 | VARIABLE l_paren args R_PAREN { $$ = 0; }
                 ;


bad_term:          VARIABLE { $$ = 0;
                   ql_parse_error
                     ("A predicate name must begin with a lower-case letter"); }
                 | var_paren_term { $$ = 0;
                   ql_parse_error
                     ("A predicate name must begin with a lower-case letter"); }
                 | ws_paren_term { $$ = 0;
                   ql_parse_error
                     ("No space allowed between predicate name and left parenthesis"); }
                 ;


good_paren_term:   TOK_STRING L_PAREN R_PAREN
                   { $$ 
                     = QL_Term::create($1,0,error_position); }
                 | TOK_STRING L_PAREN args R_PAREN
                   { $$
                     = QL_Term::create((void*)$1,(void*)$3,error_position); }
                 | TWO_OP L_PAREN args R_PAREN 
                   { $$
                     = QL_Term::create((void*)$1,(void*)$3,error_position); }
                 ;

ws_paren_term:    TOK_STRING WS_L_PAREN R_PAREN  { $$ = 0; }
                 | TOK_STRING WS_L_PAREN args R_PAREN  { $$ = 0; }
                 | TWO_OP WS_L_PAREN args R_PAREN { $$ = 0; }
                 ;

good_term:         TOK_STRING
                   { $$ 
                     = QL_Term::create((void*)$1,(void*)0,error_position); }
                 | good_paren_term { $$ = $1; }

args:            arg { $$ = $1; } 
	       | arg COMMA args { 
                    if ($1 && $3)
                       $$ = ((QL_Args*)$1)->set_next((QL_Args*)$3);
                    else $$ = 0;
                 }
               ;

arg:             simple_arg { $$ = $1; }
               | tag_assoc { $$ = new QL_Args((QL_Term*)$1); }

simple_arg:      constant { $$ = new QL_Args((Value*)$1); }
               | VARIABLE 
                 { $$ = (void *) new QL_Args(bindings->add((String *)$1)); }
               | good_paren_term 
                 { $$ = new QL_Args(/*x*/(QL_Term*)$1); }
               | var_paren_term 
                 { $$ = 0;
                   ql_parse_error
                     ("A predicate name must begin with a lower-case letter");
                 }
               /* | L_PAREN args R_PAREN { $$ = $2 } */
               ;

c_args:          c_arg { $$ = $1; } 
	       | c_arg COMMA c_args
                 { $$ = ((QL_Args*)$1)->set_next((QL_Args*)$3); }
               ;

c_arg:           constant { $$ = new QL_Args((Value*)$1); }
               | l_paren c_args R_PAREN 
                 { /* See about deleting this production ASAP */
                   $$ = new QL_Args(V_Collection::make(Value::LIST, (QL_Args*)$2)); }
               ;


collection:      list { $$ = $1; }
               | set  { $$ = $1; }
               | bag  { $$ = $1; }
               ;


list:            L_LIST R_LIST
                 { $$ = V_Collection::make(Value::LIST); }
               | L_LIST c_args R_LIST { $$ = V_Collection::make(Value::LIST, (QL_Args*)$2); }
               ;

set:             L_SET R_SET
                 { $$ = V_Collection::make(Value::SET); }
               | L_SET c_args R_SET { $$ = V_Collection::make(Value::SET, (QL_Args*)$2); }
               ;

bag:             L_BAG R_BAG
                 { $$ = 0;
                   ql_parse_error("Bags not yet supported");
                   /* $$ = new V_Collection(Value::BAG); */ }
               | L_BAG c_args R_BAG
                 { $$ = 0;
                   ql_parse_error("Bags not yet supported");
                   /* ((QL_Args*)$2)->to_BAG(); */
                 }
               ;

constant:        TOK_STRING   { $$ = $1; }
               | TOK_DATE     { $$ = $1; }
               | TOK_FLOAT    { $$ = $1; }
               | TOK_INTEGER  { $$ = $1; }
               | collection   { $$ = $1; }
               ;


tag_assoc:       simple_arg TWO_OP tag_assoc_rhs
                 { if ($1) {
                      ((QL_Args*)$1)->set_next((QL_Args*)$3);
                      $$ = QL_Term::create((void*)$2,(QL_Args *)$1,
                                           error_position);
                   } else $$ = 0;
                 }
               ;


empty:           /* empty; this rule is needed so the rule for
                   tag_assoc_rhs can return 0.  */
               ;

tag_assoc_rhs:   empty { $$ = 0; }
               | arg { $$ = $1; }
               ;
%%
