/* 
 * Copyright (c) 1995
 *       Whitehead Institute for Biomedical Research. All rights reserved.
 *
 * Released under terms specified in the Copyright file present in the top
 * directory of this distribution.
 */


#ifndef QL_DD_ENTRY_H
#define QL_DD_ENTRY_H

#ifndef TAG_H
#include "tag.H"
#endif

#ifndef BASICS_H
#include "basics.H"
#endif

#ifndef VALUE_H
#include "value.H"
#endif

// #if defined(TEXAS) && defined(USE_INDEX)
// #ifndef BTREE_HH
// #include "btree.hh"
// #endif
// #endif

#ifndef QL_BUILTIN_IMPL_H
#include "ql_builtin_impl.H"
#endif

#ifndef ID_INDEX_ENTRY_H
#include "id_index_entry.H"
#endif

class QL_Adornment_Map;
class QL_Term;
class QL_Args;
class QL_Iterator;
class QL_Iterator_State;
class Id_Index_Entry;

class QL_DD_Tag_Entry;
class QL_DD_Builtin_Entry;
class QL_DD_Material_Kind_Entry;
class QL_DD_Step_Kind_Entry;
class QL_DD_DD_Kind_Entry;
class QL_DD_Type_Constructor_Entry;

#define HARDWIRED_USER "Constructed automatically at server startup"

class QL_DD_Entry
{ 
public:
  // The enum kind is the run-time type descriptor of classes
  // derived from QL_DD_Entry.
  enum kind { TAG, BUILTIN, MATERIAL_KIND, STEP_KIND, DATA_DICTIONARY_KIND,
	      TYPE_CONSTRUCTOR };

  // Return the QL_DD_Entry::kind of `this'.
  virtual kind entry_kind() = 0;

  QL_DD_Entry(char *identifier, char *who="Not specified",
	      Write_implementation ii=0, Write_implementation di=0,
	      Write_implementation ui=0);

  // The following functions provide fail-stop downcasts to the
  // classes derived from QL_DD_Entry.  They abort if 'this'
  // QL_DD_Entry does not have the right subclass.
  QL_DD_Tag_Entry *down_to_TAG(); 
  QL_DD_Builtin_Entry *down_to_BUILTIN();
  QL_DD_Material_Kind_Entry *down_to_MATERIAL_KIND();
  QL_DD_Step_Kind_Entry *down_to_STEP_KIND();
  QL_DD_DD_Kind_Entry *down_to_DATA_DICTIONARY_KIND();
  QL_DD_Type_Constructor_Entry *down_to_TYPE_CONSTRUCTOR();

  SLOT(int,retired);
  SLOT(char*,identifier);
  SLOT(char*,who);


  // ************************** MAKE CHANGE ********************************
  //
  // The following data members should be moved down the class hierarchy, so
  // that they do not interfere with the generation of class descriptions in
  // Texas.  (The should be moved down to the leaf classes.)

  SLOT(Write_implementation,implementation_of_insert);
  SLOT(Write_implementation,implementation_of_delete);
  SLOT(Write_implementation,implementation_of_update);

protected:
  friend class DD_Ops; // For access to several protected functions.
  friend class QL_DD;

  // Go through the list headed by `this' in search of an entry
  // with identifier() `id'.
  static QL_DD_Entry *lookup(QL_DD_Entry*, char *id);

  // Pointer to the next QL_DD_Entry in the list.
  SLOT(QL_DD_Entry*,next);
protected:
  void set_next(QL_DD_Entry *next) { _next = next; }
};


class QL_DD_Builtin_Entry : public QL_DD_Entry
{
public:
  QL_DD_Builtin_Entry(char *identifier, Write_implementation ii=0,
		      Write_implementation di=0, Write_implementation ui=0);
  kind entry_kind() { return BUILTIN; }
  SLOT(QL_Adornment_Map *,adornment_map);
};


class QL_Type_Descriptor
{
public:
  QL_Type_Descriptor(Value::tv_type_id);
  QL_Type_Descriptor(QL_Type_Descriptor*);
  QL_Type_Descriptor(QL_Term*,QL_Args*);
  ~QL_Type_Descriptor();
  static QL_Type_Descriptor** find_in_args(QL_Term*,QL_Args*,int*);
  add(QL_Type_Descriptor);
  SLOT(Value::tv_type_id,constructor);
  
  // In the future we may wish to view material kinds
  // and step kinds as types.
  // No code currently depends on this slot.
  SLOT(QL_DD_Entry*,dd_constructor); 

  SLOT(int,arg_number);
  SLOT(QL_Type_Descriptor**,args);
};


class QL_DD_Tag_Entry : public QL_DD_Entry
{
public:
  kind entry_kind() { return TAG; }
  static QL_DD_Tag_Entry *
    new_dd_tag(char *identifier, Value::tv_type_id c_type, Tag tag_index);

  QL_DD_Tag_Entry(char *identifier,         // The tag name.

                  Tag tag_index,	    // The tag index itself.

                  char *who,		    // The user creating the tag.

		  QL_Type_Descriptor *type, // The type of the tag.

                  int create_only=0,	    // If non-0, then the tag can occur
                                            // only once, in the 'create' step.

                  int id_tag=0,		    // If non-0, the tag is an id tag.

		  int segment=-1,	    // If > 0, the segment index
					    // in which to create values for
					    // the tag. Not fully implemented.

                  int put_on_history=1,     // If non-0, put the steps
					    // containing this tag on the
                                            // history of materials in the 
                                            // tag's value.
                                            // Not fully implemented.

		  int associate_with_id=1   // Meaningfull only if the
                                            // type of this tag is
                                            // MATERIAL_POINTER.
                                            // If true, associate materials
                                            // associated with this tag 
                                            // with any id tags in the same
					    // step. NOT COMPLETELY IMPLEMENTED.
		  );

  static char* primary_id_name(char *);

  int check_duplicate(QL_Term *, char *id);
  // Call ql_eval_error and return 0 if id is already
  // in the id_index.  Otherwise return 1.

  const Id_Index_Entry *lookup_material(QL_Term*, char*);  

  SLOT(int,dd_tag);
  SLOT(int,create_only);
  SLOT(int,id_tag);
  SLOT(int,segment_id);
  SLOT(int,put_on_history);
  SLOT(Tag,tag_index);
public: 
  // We cannot use the SLOT macro because ID_INDEX_SET expands to
  // BTtree<Str_Ptr,Material_Ptr>, and the comma messes up the macro.
  ID_INDEX_SET *id_index() const{ return _id_index;}
protected:
  ID_INDEX_SET *_id_index; // One index for all materials with which this
                           // tag is associated n some Step, meaningful if
                           // id_tag is true.
  SLOT(QL_Type_Descriptor*,type);
  SLOT(int,associate_with_id);
};


class QL_DD_Material_Kind_Entry : public QL_DD_Entry
{
public:
  QL_DD_Material_Kind_Entry(char *identifier, char *who, ID_INDEX_SET*);
  SLOT(ID_INDEX_SET*,main_index); // For the index on "<kind>_id".  Also is
                                  // the extent of materials of a paraticlar
                                  // kind.

  // In case we ever want is-a relations among materials.
  // No code depends on this slot.
  SLOT(QL_DD_Material_Kind_Entry*,supertype);
  kind entry_kind();
};  


class QL_DD_Step_Kind_Entry : public QL_DD_Entry
{
public:
  QL_DD_Step_Kind_Entry(char *identifier, char *who);

  // In case we ever want is-a relations among steps
  // No code depends on this slot.
  SLOT(QL_DD_Step_Kind_Entry*,supertype);

  kind entry_kind() { return STEP_KIND; }
};


class QL_DD_DD_Kind_Entry : public QL_DD_Entry
{
public:
  QL_DD_DD_Kind_Entry(char *identifier, Write_implementation,
		      Write_implementation di=0, Write_implementation ui=0);
  kind entry_kind() { return DATA_DICTIONARY_KIND; }
};


class QL_DD_Type_Constructor_Entry : public QL_DD_Entry
{
public:
  QL_DD_Type_Constructor_Entry(char *identifier,
			       Value::tv_type_id type,
			       int has_args);
  kind entry_kind() { return TYPE_CONSTRUCTOR; }
  SLOT(int,has_args);
  SLOT(Value::tv_type_id,constructor);

  // No code depends on the following slots.

  // In case we ever have constructed types:
  SLOT(QL_Type_Descriptor*,named_type);
  // Probably will not be useful:
  SLOT(QL_DD_Entry*,supertype);

};
#endif
