/* 
 * Copyright (c) 1995
 *       Whitehead Institute for Biomedical Research. All rights reserved.
 *
 * Released under terms specified in the Copyright file present in the top
 * directory of this distribution.
 */


// $Id: ql_term.C,v 1.5 1995/02/13 17:50:53 steve Exp $
#include "ql_term.H"

#include <string.h>

#ifndef QL_ARGS_H
#include "ql_args.H"
#endif

#ifndef QL_DD_H
#include "ql_dd.H"
#endif

#ifndef QL_DD_ENTRY_H
#include "ql_dd_entry.H"
#endif

#ifndef BASICS_H
#include "basics.H"
#endif

#ifndef VALUE_H
#include "value.H"
#endif

#ifndef QL_ADORNMENT_MAP_H
#include "ql_adornment_map.H"
#endif

#ifndef QL_ERRROR_H
#include "ql_error.H"
#endif

#ifndef QL_BUILTIN_IMPL_H
#include "ql_builtin_impl.H"
#endif


QL_GC(QL_Term) // Defined in ql_gc.H


void *QL_Term::create(void *void_predicate, void*void_args,
		      QL_Position& parse_position)
{
  // See discussion elsewhere of rationale for using
  // void *s in the bison-generated parser.
  String *predicate = (String *)void_predicate;
  QL_Args *args = (QL_Args *)void_args;
  char *id = predicate->get();
  QL_DD_Entry *e = QL_DD::find_or_error(id);
  delete id;
  if (!e) return 0;
  if (!strcmp(e->identifier(),"="))
    return new QL_Tag_Assoc(e,args,parse_position);
  else
    return new QL_Term(e,args,parse_position);  // We do not know the binding
}


QL_Term::QL_Term(QL_DD_Entry *dd_entry, QL_Args *args, QL_Position position) :
    Value(TERM),
    _dd_entry(dd_entry),_implementation(0),_parse_position(position),
    _args(args),_next(0),_garbage_next(_garbage_list),
    _number_free(-1)
      { _garbage_list = this; }


void QL_Term::set_adornment_info()
{
  ABORT(!_dd_entry,"No dd entry");
  _adornment = _args->adornment();
  _number_free = _adornment.number_free();
  _number_bound = _adornment.number_bound();
}
    

int QL_Term::set_implementation()
{
  ABORT(!_dd_entry,"No dd entry");
  ABORT(_dd_entry->entry_kind() != QL_DD_Entry::BUILTIN, "Need a builtin");
  _implementation
    = ((QL_DD_Builtin_Entry*)_dd_entry)->adornment_map()->find(_adornment);
  if (!_implementation) {
    ql_eval_error(this, "Adornment (binding pattern) ",
		  _adornment.braces(), " is unknown for");
    return 0;
  }
  return 1;
}


void QL_Term::print(ostream& os) const {  os << "(a term)"; }


QL_Term *QL_Term::save(P_SEGMENT*) const { ABORT(1,"Cannot yet 'save' a term"); return 0; }


int QL_Term::compare(const Value& tv)
{
  if (tv.tv_type() == TERM) return ((QL_Term*)&tv) == this;
  else { set_error("Attempt to compare a 'TERM' to a non-'TERM'");  return 1; }
}


QL_Tag_Assoc::QL_Tag_Assoc(QL_DD_Entry *dd_entry, QL_Args *args, QL_Position position)
: QL_Term(dd_entry, args ? args->next(): 0, position)
{
  if (!args || !args->next())
    {  ql_parse_error("A tag association must have arguments");
       return;
     }

  char *optr = dd_entry->identifier();
  if (*optr == '=' && *(optr + 1) == '\0') _op = EQUAL;
  else ABORT(1,"Illegal tag association operator");

  Value *tv = args->value();
  if (!tv || tv->tv_type() != Value::STRING) {
    ql_parse_error("Must have a constant string as a tag");
    return;
  }

  char *c_tag = tv->down_to_STRING()->get();
  _tag = QL_DD::tag_find(c_tag);
  if (!_tag) {
    ql_parse_error("'", c_tag, "' is not a known tag");
    delete c_tag;
    return;
  }
  delete c_tag;
#ifdef FINISHED
Add code here to recusively check arity of sub-values (?)
#endif    
//  if (_args && !_tag->variadic() && _tag->number_types() != _args->arity())
//
//      ql_parse_error("Incorrect number of values for tag '",
//		     c_tag, "'");

}


// Might want to add a check_type_if_bound function
// that doesn't complain about unbound variables.
// Such a function could check type at parse time in many cases.

Value *QL_Tag_Assoc::check_or_coerce_type(QL_Term *t)
{
  Value *v = args()->value();
  char *tag_id = _tag->identifier();
  if (!v) {
    ql_eval_error(t, "Unbound variable associated with tag '", tag_id, "' in");
    return 0;
  }
  return check_type_1_tag_value(t, tag_id, v, _tag->type());
}


Value *QL_Tag_Assoc::check_type_1_tag_value(QL_Term *t, char* tag_id, Value *tv,
					 const QL_Type_Descriptor *type)
{
  switch (type->constructor()) {
  case Value::SET: 
  case Value::LIST: 
  case Value::SLIST:
  case Value::TUPLE:
    return check_type_collection(t,tag_id, tv, type);
  case Value::BOOLEAN: return check_type_boolean(t,tag_id,tv);
  case Value::INTEGER: return check_type_integer(t,tag_id,tv);
  case Value::DNA_SEQUENCE: return check_type_dna_sequence(t,tag_id,tv);
  default: if (check_tag_value_type(t,&tv,type->constructor(),tag_id)) 
    return tv; else return 0;
  }
}


void tag_type_error(const QL_Term *t, Value::tv_type_id type,
		    const char *r1, const char *r2, const char *r3)
{
  const char *msg = Value::type2char(type);;
  // if (Value::INTEGER32 == type) msg = "'INTEGER' or 'INTEGER32'";
  // else msg = Value::type2char(type);
  ql_eval_error(t, "Expected value of type '", msg, "' ", r1, r2, r3, " in");
}


int check_tag_value_type(QL_Term *t,
			 Value **v, Value::tv_type_id type, char *tag_id)
{
  // if (Value::INTEGER == type) type = Value::INTEGER32;
  if (type_ok(t, v, type)) return 1;
  if (type == Value::TERM)
    ql_eval_error(t, "Expected value of type term to be associated with tag '",
		  tag_id, "' in");
  else tag_type_error(t, type, "to be associated with tag '",  tag_id, "'");
  return 0;
}


Int *
QL_Tag_Assoc::check_type_boolean(QL_Term *t, char* tag_id, Value *tv)
{
  Int *ret  = check_type_integer(t, tag_id, tv);
  if (!ret) return 0;
  if (1 != int(*ret) && 0 != int(*ret)) {
    ql_eval_error(t, "Expected 1 or 0 to be associated with tag '",
		  tag_id, "' in");
    return 0;
  }
  return ret;
}


Int *
QL_Tag_Assoc::check_type_integer(QL_Term *t, char* tag_id, Value *v)
{
  if (!check_tag_value_type(t,&v,Value::INTEGER,tag_id)) return 0;
  else return v->down_to_INTEGER();
//   int x = int(*v->down_to_INTEGER());
//   clear_error();
//   Int *ret = Int::make(x);
//   if (errmsg()) {
//     ql_eval_error(t, errmsg(), " in");
//     return 0;
//   }
//   return ret;
}


Value *QL_Tag_Assoc::check_type_dna_sequence(QL_Term *t, char* tag_id, Value *tv)
{
  if (tv->tv_type() == Value::DNA_SEQUENCE) return tv;
  if (!check_tag_value_type(t,&tv,Value::STRING,tag_id)) return 0;
  Sequence *seq = Sequence::make(tv->down_to_STRING());
  if (seq->valid()) return seq;
  char *sequence = seq->get();
  ql_eval_error(t, "Invalid nucleotide sequence '", sequence, "'");
  delete sequence;
  return 0;
}


Value *QL_Tag_Assoc::check_type_collection(QL_Term *t, char* tag_id,
					Value *v,
					const QL_Type_Descriptor *type)
{
  tv_type_id constructor = type->constructor();
  ABORT(constructor != Value::LIST && constructor != Value::SLIST
	&& constructor != Value::SET && constructor != Value::TUPLE,
	"Needed a collection");

  if ((constructor == Value::LIST || constructor == Value::SET)
      && !check_tag_value_type(t,&v,type->constructor(),tag_id)) return 0;
  const V_Collection *collection;
  switch (v->tv_type()) {
  case Value::LIST:
    collection = v->down_to_LIST(); break;
  case Value::SET:
    collection = v->down_to_SET(); break;
  case Value::SLIST:
    collection = v->down_to_SLIST(); break;
  default: 
    check_tag_value_type(t,&v,type->constructor(),tag_id); return 0;
  }
  V_Cursor* c = collection->cursor_of();
  
  Value *tv, *xv1;
  int i, index = 0;
  for (i = 0, tv = c->first(); tv; tv = c->next(), i++) {

    if (constructor == Value::TUPLE) {
      if (i >= type->arg_number()) {
	ql_eval_error(t, "Too many values in tuple in");
	delete c; return 0;
      }
      index = i;
    }

    xv1 = check_type_1_tag_value(t, tag_id, tv, type->args()[index]);
    if (!xv1) {
      ql_eval_error(t, "Collection component with illegal type in");
      delete c;
      return 0;
    }
    // Need to add code to do the coercion!
    if (xv1 != tv) {
      ql_eval_error(t, "Cannot coerce type of nested values in");
      delete c;
      return 0;
    }
  }
  delete c;

  if (i < type->arg_number() && constructor == Value::TUPLE) {
    ql_eval_error(t, "Too few values in tuple in");
    return 0;
  }

  //  OLD CODE from before the time that there was an explicit TUPLE constructor:
  //
  //       if (tv->tv_type() != Value::LIST) {
  //	 ql_eval_error(t, "Collection component with illegal type in");
  //	 delete c;
  //	 return 0;
  //       }
  //       V_Cursor*cc = ((V_Collection*)tv)->cursor_of();
  //       int i = 0;
  //       for (Value *tv2 = cc->first(); tv2; tv2 = cc->next(), i++) {
  //	 if (i > type->arg_number()) {
  //	   ql_eval_error(t, "Collection component with too many values in");
  //	   delete cc; delete c;
  //	   return 0;
  //	 }
  //	 Value *xv2 = check_type_1_tag_value(t, tag_id, tv2, type->args()[i]);
  //	 if (!xv2) { delete c; delete cc; return 0; }
  //	 if (xv2 != tv2) {
  //	   ql_eval_error(t, "Cannot coerce type of nested values in");
  //	   delete c; delete cc; return 0;
  //	 }
  //       }
  //       delete cc;
  //       if (i < type->arg_number()) {
  //	 ql_eval_error(t, "Collection component with too few values in");
  //	 delete c;
  //	 return 0;
  //       }
  
  switch (constructor) {
  case Value::TUPLE:
    if (v->tv_type() != Value::LIST) {
      tag_type_error(t, Value::TUPLE, 0, 0, 0); return 0;
    } else return v;
  case Value::LIST: case Value::SET:
    return v;
  case Value::SLIST:
    if (collection->tv_type() == Value::SLIST) return v;
    ABORT(type->arg_number() != 1,"Require 1 type argument");
    tv_type_id arg_constructor = type->args()[0]->constructor();
    Value *zero = Value::zero_of(arg_constructor);
    if (!zero) {
      ql_eval_error(t, "Cannot create an SLIST with subtype ", "", "in");
      return 0;
    } else return V_Collection::make(Value::SLIST, collection, zero);
  default: ABORT(1, "Illegal type/programming error");
  }
}
