/* 
 * Copyright (c) 1995
 *       Whitehead Institute for Biomedical Research. All rights reserved.
 *
 * Released under terms specified in the Copyright file present in the top
 * directory of this distribution.
 */


// $Id: step_query.C,v 1.4 1995/02/13 17:51:07 steve Exp $
#ifndef	__sys_limits_h
#include <limits.h>
#endif

#ifndef STEP_QUERY_H
#include "step_query.H"
#endif

#ifndef BASICS_H
#include "basics.H"
#endif

#ifndef VALUE_H
#include "value.H"
#endif

#ifndef STEP_H
#include "step.H"
#endif

#ifndef MATERIAL_H
#include "material.H"
#endif

#ifndef QL_TERM_H
#include "ql_term.H"
#endif

#ifndef QL_ARGS_H
#include "ql_args.H"
#endif


#ifndef QL_DD_ENTRY_H
#include "ql_dd_entry.H"
#endif

#ifndef QL_BUILTIN_IMPL_H
#include "ql_builtin_impl.H"
#endif

#ifndef QL_ITERATOR_H
#include "ql_iterator.H"
#endif

#ifndef QL_EVAL_H
#include "ql_eval.H"
#endif

#define A_SIZE 100

class Step_Query
// Step_Query is quite similar to QL_Iterator, but doesn't
// seem to be logically a subclass.
{
public:
  Step_Query(QL_Term*);
  ~Step_Query();
  static QL_FIRST_HEADER(first);
  static QL_NEXT_HEADER(next);

private:
  int setup_1_arg(QL_Term*,QL_Args*);

  int matches(Step*);
  void clear_bindings();

  QL_DD_Step_Kind_Entry *step_kind;
  // We will need the step kind to know which steps to consider.

  QL_Term *t;

  int ok; // Set to 0 if this is not usable.
  int more; // Set to 0 if there are no more results.

  int check_size(int s);

  int m_input_number, m_output_number, o_input_number,
      o_output_number;

  Material *m_inputs[A_SIZE];
  Value *o_inputs[A_SIZE];
  QL_Binding_Node *m_nodes[A_SIZE], *o_nodes[A_SIZE];
  QL_DD_Tag_Entry *m_inputs_tags[A_SIZE], *m_outputs_tags[A_SIZE],
                  *o_inputs_tags[A_SIZE], *o_outputs_tags[A_SIZE];


  List_Material *m_set;  // This should be generalized to a set of _tuples_
			  // of materials!

  QL_Iterator_State *find_match();
  Material *driving_material;
  Cursor_Step *cursor;
};

int Step_Query::check_size(int s)
{
  if (s >= A_SIZE) {
    ql_eval_error(t, "Limit exceeded (see file '", __FILE__, "') in");
    return 0;
  } else return 1;
}


Step_Query::Step_Query(QL_Term *t)
: m_output_number(0), m_input_number(0),  o_output_number(0),
  o_input_number(0), ok(0), driving_material(0), cursor(0),  m_set(0), more(1)
{
  step_kind = t->dd_entry()->down_to_STEP_KIND();
  // Each of the arguments should bound and should
  // be a tag association.
  // However, not all the tag association operands
  // need to be constants.  There must, however,
  // be at least one with a MATERIAL_POINTER operand.

  // Add code to find the material(s).
  for (QL_Args *p = t->args(); p; p = p->next())
    if (!setup_1_arg(t, p)) return;

  if (!m_input_number)
    ql_eval_error(t, "At least one material tag must be bound in");
  else {
    m_set = new List_Material();
    ok = 1;
  }
}


int Step_Query::setup_1_arg(QL_Term *t, QL_Args *p)
{
  QL_Tag_Assoc *ta = p->coerce_to_Tag_Assoc();
  ABORT(!ta,"This should have been caught upstream");

  Value *tv = ta->args()->value();
  Value::tv_type_id tv_type = ta->tag()->type()->constructor();

  if (tv && !check_tag_value_type(t, &tv, tv_type, ta->tag()->identifier()))
    return 0;

  if (Value::MATERIAL_POINTER == tv_type) {
    if (tv) {
      if (!check_size(m_input_number)) return 0;
      m_inputs[m_input_number] = tv->down_to_MATERIAL_POINTER();
      m_inputs_tags[m_input_number] = ta->tag();
      m_input_number++;
    } else {
      if (!check_size(m_output_number)) return 0;
      m_nodes[m_output_number] = ta->args()->node();
      ABORT(!m_nodes[m_output_number], "Invalid QL_Arg");
      m_outputs_tags[m_output_number] = ta->tag();
      m_output_number++;
      if (m_output_number > 1) {
	ql_eval_error(t, "No more than 1 unbound material is allowed in");
	return 0;
      }
    }
  } else { // Not a material pointer
    if (tv) {
      if (!check_size(o_input_number)) return 0;
      o_inputs[o_input_number] = tv;
      o_inputs_tags[o_input_number] = ta->tag();
      o_input_number++;
    } else {
      if (!check_size(o_output_number)) return 0;
      o_nodes[o_output_number] = ta->args()->node();
      ABORT(!o_nodes[o_output_number], "Invalid QL_Arg");
      o_outputs_tags[o_output_number] = ta->tag();
      o_output_number++;
    }
  }
  return 1;
}


Step_Query::~Step_Query()
{
  clear_bindings();
  delete m_set;
  delete cursor;
}


void Step_Query::clear_bindings()
{
  int i;
  for (i = 0; i < m_output_number; i++)  m_nodes[i]->add_value(0);
  for (i = 0; i < o_output_number; i++) o_nodes[i]->add_value(0);
}


QL_FIRST_HEADER(Step_Query::first)
{
  if (in || out || m || n); // To suppress compiler warnings.
  if (inside_exists) {
    ql_eval_error(t, "Cannot evaluate a step query inside the 'exists' predicate in");
    return 0;
  }
  Step_Query *r = new Step_Query(t);
  if (r->ok) {
    QL_Iterator_State *s;
    if (s = r->find_match()) return s;
    else { delete r; return 0; }
    // find_match picks material with fewest steps, then
    // goes through its steps looking for a match.
  }
  else { delete r; return 0; }
}


QL_NEXT_HEADER(Step_Query::next)
{
  if (in || out) m = n = 0; t = 0; // To suppress compiler warnings.
  if (((Step_Query*)s)->find_match()) return s;
  else { delete (Step_Query*) s; return 0; }
}


QL_Iterator_State *Step_Query::find_match()
{
  // Create the cursor if necessary.
  Step *s;
  if (cursor) s = cursor->previous();
  else {
    int smallest = INT_MAX;
    {
      for (int i = 0; i < m_input_number; i++) {
        Material *m = m_inputs[i];
        int card =  m->history()->count();
	if (card < smallest) { smallest = card; driving_material = m; }
      }
    }
    ABORT(!driving_material,"Invariant violated");
    cursor = driving_material->history()->cursor_of();
    s = cursor->last();
  }

  if (!more) return 0;


  while (s) {
    if (matches(s)) return (QL_Iterator_State*) this;
    // The bindings are set up in matches.
    if (!more) return 0;
    s = cursor->previous();
  }
  return 0;
}


int Step_Query::matches(Step* s)
{
  // It is not necessary to clear the output bindings
  // at the start of matches, because once the 
  // iterator is set up it knows which arguments
  // are bound and which are free.

  int i;
  if (s->step_kind() != step_kind) return 0;
  
  // Check that each of the bound materials (other than the
  // driving material) matches.

  for (i = 0; i < m_input_number; i++)
    if (m_inputs[i] != driving_material) {
      Value *v = s->find(m_inputs_tags[i]->tag_index());
      if (!v || (v != m_inputs[i]))
	return 0;
    }

  // For each tuple of unbound materials in the step, add it to the
  // corresponding set of "considered material tuples" (m_set).  If any 
  // requested material is absent, this step cannot match.

  for (i = 0; i < m_output_number; i++) {
    Value *v = s->find(m_outputs_tags[i]->tag_index());
    if (!v) return 0;
    Material *m = v->down_to_MATERIAL_POINTER();
    if (m_set->contains(m)) return 0;
    m_set->put(m);
    m_nodes[i]->add_value(v);
  }

  // For each bound other, check that it agrees with value for the tag in s.

  for (i = 0; i < o_input_number; i++) {
    Value *v = s->find(o_inputs_tags[i]->tag_index());
    if (!v) { if (0 == m_output_number) more = 0; return 0; }
    if (*o_inputs[i] != *v) { if (0 == m_output_number) more = 0; return 0; }
  }

  // Set up the bindings for the other outputs.

  for (i = 0; i < o_output_number; i++) {
    Value *v = s->find(o_outputs_tags[i]->tag_index());
    if (!v) return 0;
    o_nodes[i]->add_value(v);
  }

  if (m_output_number == 0) more = 0;
  return 1;
}


static QL_FIRST_HEADER(test_step_kind)
// This function implements predicates of the form
//
// <step_kind>(S)`
//
//  where S is a step (or bound to a step).
//  It expects term to be of arity 1 with it's
//  single argument bound to a Value.
{
  abort_m_n(1,m,0,n);
  if (out); // Suppress warnings.
  if (in[0]->tv_type() != Value::STEP_POINTER) {
    ql_eval_error(t, "Need an argument of type 'STEP_POINTER' for");
    return 0;
  }
  if (in[0]->down_to_STEP_POINTER()->step_kind() == t->dd_entry())
    return (QL_Iterator_State*) 1;
  else return 0;
}


QL_Iterator *create_step_kind_iterator(QL_Term*t)
{
  {
    Value *tv;
    if (t->args()->arity() == 1 && (tv = t->args()->value())
	&& tv->tv_type() != Value::TERM)
      return new QL_Iterator(t, 1, 0, test_step_kind, 0);
  }
  int bound_number = 0;
  for (QL_Args *p = t->args(); p; p = p->next()) {
    QL_Tag_Assoc *ta = p->coerce_to_Tag_Assoc();
    if (!ta) { ql_eval_error(t,"All arguments must be tag associations for");
	       return 0; 
	     }
    bound_number++;
  }
  // The rest of the initial error checking occus in
  // step_kind_first, which ignores m and n
  // given in the constructor for QL_Iterator
  // (whose default method for managing
  // bound and unbound arguments will not for
  // step kind queries).

  return 
    new QL_Iterator(t, bound_number, 0, Step_Query::first, Step_Query::next);
}
